@@ -61,9 +61,10 @@ const CHUNK_CONTENT_TYPE: &str = "application/octet-stream";
 /// We want to guarantee a low average rate (e.g., 30 per second)
 /// to bound our costs, but also avoid slowing down replication in
 /// the common case, when write transactions are rare.
+const COPY_RATE: NonZeroU32 = unsafe { NonZeroU32::new_unchecked(30) };
+
 const COPY_RATE_QUOTA: governor::Quota =
-    governor::Quota::per_second(unsafe { NonZeroU32::new_unchecked(30) })
-        .allow_burst(unsafe { NonZeroU32::new_unchecked(100) });
+    governor::Quota::per_second(COPY_RATE).allow_burst(unsafe { NonZeroU32::new_unchecked(100) });
 
 /// Rate limit when copying synchronously.
 ///
@@ -93,12 +94,6 @@ const COPY_RETRY_BASE_WAIT: Duration = Duration::from_millis(100);
 /// consecutive failures.
 const COPY_RETRY_MULTIPLIER: f64 = 10.0;
 
-/// Remember up to this many recent copy requests for deduplication.
-///
-/// This value should ideally be proportional to the maximum copy rate
-/// (`COPY_RATE_QUOTA`) and `COPY_REQUEST_MIN_AGE`.
-const COPY_REQUEST_MEMORY: usize = 30 * 3600;
-
 /// Avoid executing duplicate copy requests that are separated by less
 /// than `COPY_REQUEST_MIN_AGE`.
 const COPY_REQUEST_MIN_AGE: Duration = Duration::from_secs(3600);
@@ -107,6 +102,17 @@ const COPY_REQUEST_MIN_AGE: Duration = Duration::from_secs(3600);
 /// timestamp by up to this duration.
 const COPY_REQUEST_JITTER: Duration = Duration::from_secs(600);
 
+/// Remember up to this many recent copy requests for deduplication.
+///
+/// We would like to scale `COPY_REQUEST_MIN_AGE /
+/// COPY_RATE_QUOTA.replenish_interval()`, but `const` isn't there
+/// yet.  The memory size is a bit larger than the number of uploads
+/// we expect during `COPY_REQUEST_MIN_AGE` to avoid spuriously losing
+/// useful entries to LRU... but we still want a bound on the capacity,
+/// because unbounded data structures are a bad idea.
+const COPY_REQUEST_MEMORY: usize =
+    (1.5 * (COPY_RATE.get() as f64) * (COPY_REQUEST_MIN_AGE.as_secs() as f64)) as usize;
+
 /// Perform background work for one spooling directory approximately
 /// once per BACKGROUND_SCAN_PERIOD.
 const BACKGROUND_SCAN_PERIOD: Duration = Duration::from_secs(5);
@@ -190,12 +196,19 @@ const REPLICATION_LAG_REPORT_THRESHOLD: Duration = Duration::from_secs(120);
 /// Time background "touch" to fully cover each db file once per period.
 const PATROL_TOUCH_PERIOD: Duration = Duration::from_secs(24 * 3600);
 
-/// Compression level for zstd. 0 maps to zstd's internal default
-/// compression level.
-const ZSTD_COMPRESSION_LEVEL: i32 = 0;
+/// Zstd compression level for chunk data. 0 maps to zstd's internal
+/// default compression level.
+const CHUNK_COMPRESSION_LEVEL: i32 = 0;
 
-/// Disable compression with a negative level value.
-const ZSTD_COMPRESSION_DISABLE: i32 = -1;
+/// Fast compression level.  We use this zstd level for manifests
+/// because we expect data to either be incompressible, or feature
+/// long spans of trivially compressible 0s.
+const FAST_COMPRESSION_LEVEL: i32 = 1;
+
+/// When the raw data is at least this large, convert default (0)
+/// compression to `FAST_COMPRESSION_LEVEL`: we assume that's not
+/// sqlite db data, and probably incompressible.
+const FAST_COMPRESSION_AUTO_SIZE: usize = (crate::tracker::SNAPSHOT_GRANULARITY as usize) + 1;
 
 lazy_static::lazy_static! {
     static ref RECENT_WORK: Mutex<RecentWorkSet> = Mutex::new(RecentWorkSet::new(COPY_REQUEST_MEMORY, COPY_REQUEST_JITTER));
@@ -293,6 +306,17 @@ struct CopierSpoolState {
     // Copier workers attempt to acquire this mutex before processing
     // the spool path's contents.
     upload_lock: Mutex<CopierUploadState>,
+
+    // A shared mutable cell to retain the most recent base chunk for
+    // this db's manifest.  Keeping this chunk alive guarantees we can
+    // find it in the global cache, and thus avoids useless GETs.
+    //
+    // We `Arc` the `Mutex` because we want to capture a reference to
+    // the `Mutex` in a closure, and the lifetime gets complex.  The
+    // inner value, `Arc<Chunk>`, is refcounted because that's how the
+    // loader subsystem manages its cache, but should be treated as an
+    // opaque token that we only keep around to delay its `Drop::drop`.
+    recent_base_chunk: Arc<Mutex<Option<Arc<crate::loader::Chunk>>>>,
 }
 
 type DateTime = chrono::DateTime<chrono::Utc>;
@@ -806,6 +830,17 @@ async fn copy_file(
         .map_err(|e| chain_error!(e, "failed to read file contents", ?blob_name))?;
 
     if level >= 0 {
+        // `level == 0` tells zstd to use its internal default
+        // compression level.  If the input bytes are larger than the
+        // maximum we expect for a db chunk, assume it's probably
+        // fingerprints (incompressible), and override that by telling
+        // zstd to optimise for speed over quality.
+        let level = if level == 0 && bytes.len() >= FAST_COMPRESSION_AUTO_SIZE {
+            FAST_COMPRESSION_LEVEL
+        } else {
+            level
+        };
+
         match zstd::encode_all(bytes.as_slice(), level) {
             Ok(encoded) => bytes = encoded,
             Err(e) => tracing::warn!(?e, "failed to zstd-compress data"),
@@ -1292,7 +1327,7 @@ impl CopierWorker {
                     let chunks_buckets = chunks_buckets.clone();
 
                     Ok(Some(async move {
-                        copy_file(&name, &mut file, ZSTD_COMPRESSION_LEVEL, &chunks_buckets)
+                        copy_file(&name, &mut file, CHUNK_COMPRESSION_LEVEL, &chunks_buckets)
                             .await?;
                         RECENT_WORK.lock().unwrap().observe(&work_unit);
                         Ok(())
@@ -1331,8 +1366,7 @@ impl CopierWorker {
                     let parent = parent.to_owned();
 
                     Ok(Some(async move {
-                        copy_file(&name, &mut file, ZSTD_COMPRESSION_DISABLE, &meta_buckets)
-                            .await?;
+                        copy_file(&name, &mut file, FAST_COMPRESSION_LEVEL, &meta_buckets).await?;
                         replication_buffer::tap_manifest_file(&parent, &name, &mut file).map_err(
                             |e| {
                                 chain_warn!(
@@ -1467,7 +1501,7 @@ impl CopierWorker {
                     let chunks_buckets = chunks_buckets.clone();
                     let force_meta = force_meta.clone();
                     Ok(Some(async move {
-                        copy_file(&name, &mut file, ZSTD_COMPRESSION_LEVEL, &chunks_buckets)
+                        copy_file(&name, &mut file, CHUNK_COMPRESSION_LEVEL, &chunks_buckets)
                             .await?;
                         // Always upload the contents of the meta
                         // directory if we found chunks to upload.
@@ -1579,8 +1613,7 @@ impl CopierWorker {
                     let recent_staged_directories = state.recent_staged_directories.clone();
 
                     Ok(Some(async move {
-                        copy_file(&name, &mut file, ZSTD_COMPRESSION_DISABLE, &meta_buckets)
-                            .await?;
+                        copy_file(&name, &mut file, FAST_COMPRESSION_LEVEL, &meta_buckets).await?;
                         replication_buffer::tap_manifest_file(&parent, &name, &mut file).map_err(
                             |e| {
                                 chain_warn!(
@@ -1758,7 +1791,7 @@ impl CopierWorker {
         Ok(did_something)
     }
 
-    /// Attempts to touch a small pseudrandom subset of the chunks
+    /// Attempts to touch a small pseudorandom subset of the chunks
     /// referred by the latest uploaded manifest.
     ///
     /// Only errors out if we successfully connected to remote storage
@@ -1768,6 +1801,7 @@ impl CopierWorker {
         &self,
         spool_path: &Path,
         source: &Path,
+        recent_base_chunk: Arc<Mutex<Option<Arc<crate::loader::Chunk>>>>,
         time_since_last_patrol: Duration,
     ) -> Result<()> {
         use rand::seq::SliceRandom;
@@ -1780,32 +1814,6 @@ impl CopierWorker {
             return Ok(());
         }
 
-        // Compute what fraction of the file's current chunks we want to touch.
-        let coverage_fraction = (time_since_last_patrol.as_secs_f64()
-            / PATROL_TOUCH_PERIOD.as_secs_f64())
-        .clamp(0.0, 1.0);
-        let tap_file = replication_buffer::construct_tapped_manifest_path(spool_path, source)?;
-        let mut chunks = parse_manifest_chunks(&tap_file)?;
-        let mut rng = rand::thread_rng();
-
-        // Touch that fraction of the chunks list, with randomised
-        // rounding for any fractional number of chunks: rounding
-        // down might consistently round to 0, and rounding up
-        // ends up triggering a lot of useless work.
-        let desired_touch_count = chunks.len() as f64 * coverage_fraction;
-        // Bump the randomised rounding probability up a little: we'd
-        // rather touch too many chunks than too few, if something goes
-        // wrong with the PRNG.
-        let round_up_probability = (desired_touch_count.fract() * 2.0).clamp(0.0, 1.0);
-        let touch_count = (desired_touch_count.floor() as usize)
-            .saturating_add(rng.gen_bool(round_up_probability) as usize)
-            .clamp(0, chunks.len());
-
-        let (shuffled, _) = chunks.partial_shuffle(&mut rng, touch_count);
-        if shuffled.is_empty() {
-            return Ok(());
-        }
-
         let targets: ReplicationTargetList = {
             let metadata = replication_buffer::buffer_metadata_file(spool_path.to_path_buf());
 
@@ -1828,6 +1836,40 @@ impl CopierWorker {
             }
         };
 
+        // Compute what fraction of the file's current chunks we want to touch.
+        let coverage_fraction = (time_since_last_patrol.as_secs_f64()
+            / PATROL_TOUCH_PERIOD.as_secs_f64())
+        .clamp(0.0, 1.0);
+        let tap_file = replication_buffer::construct_tapped_manifest_path(spool_path, source)?;
+        let (mut chunks, base) = parse_manifest_chunks(&tap_file, &targets.replication_targets)?;
+        let mut rng = rand::thread_rng();
+
+        if let Some(base) = base.as_ref() {
+            chunks.push(base.fprint());
+        }
+
+        // And always overwrite the base chunk with the one we just
+        // found (or `None`).
+        *recent_base_chunk.lock().unwrap() = base;
+
+        // Touch that fraction of the chunks list, with randomised
+        // rounding for any fractional number of chunks: rounding
+        // down might consistently round to 0, and rounding up
+        // ends up triggering a lot of useless work.
+        let desired_touch_count = chunks.len() as f64 * coverage_fraction;
+        // Bump the randomised rounding probability up a little: we'd
+        // rather touch too many chunks than too few, if something goes
+        // wrong with the PRNG.
+        let round_up_probability = (desired_touch_count.fract() * 2.0).clamp(0.0, 1.0);
+        let touch_count = (desired_touch_count.floor() as usize)
+            .saturating_add(rng.gen_bool(round_up_probability) as usize)
+            .clamp(0, chunks.len());
+
+        let (shuffled, _) = chunks.partial_shuffle(&mut rng, touch_count);
+        if shuffled.is_empty() {
+            return Ok(());
+        }
+
         // Similarly, a failure here shouldn't trigger a full snapshot.
         let creds = match Credentials::default() {
             Ok(creds) => creds,
@@ -1932,11 +1974,13 @@ impl CopierWorker {
             if last_scanned > SystemTime::UNIX_EPOCH {
                 let spool_path = state.spool_path.clone();
                 let source = state.source.clone();
+                let recent_base_chunk = state.recent_base_chunk.clone();
 
                 Some(move |this: &CopierWorker| {
                     if let Err(e) = this.patrol_touch_chunks(
                         &spool_path,
                         &source,
+                        recent_base_chunk,
                         SystemTime::now()
                             .duration_since(last_scanned)
                             .unwrap_or_default(),
