@@ -26,6 +26,7 @@ use crate::replication_target::parse_s3_region_specification;
 use crate::replication_target::ReplicationTarget;
 use crate::replication_target::S3ReplicationTarget;
 use crate::result::Result;
+use crate::unzstd::try_to_unzstd;
 
 /// Writers clean up stale manifests in `.tap` on startup, but could
 /// miss some.  We also try to touch these manifest files fairly
@@ -59,18 +60,24 @@ const LOADER_POOL_SIZE: usize = 10;
 
 /// Size limit for chunks after decompression.  We expect
 /// all real chunks to be strictly shorter.
-const DECODED_CHUNK_SIZE_LIMIT: usize = 1usize << 20;
-
-/// A zstd frame starts with 0xFD2FB528 in 4 little-endian bytes.
-const ZSTD_MAGIC: [u8; 4] = [0x28, 0xB5, 0x2F, 0xFD];
+///
+/// Even a base chunk for a 1TB database only needs 256 MB (2**28
+/// bytes) for its references to data chunks.
+const DECODED_CHUNK_SIZE_LIMIT: usize = 3usize << 27;
 
 #[derive(Debug)]
-pub(crate) struct Chunk {
-    pub fprint: Fingerprint,
+pub struct Chunk {
+    fprint: Fingerprint,
     payload: memmap2::Mmap,
 }
 
 impl Chunk {
+    /// Returns the fingerprint for this chunk's data.
+    pub fn fprint(&self) -> Fingerprint {
+        self.fprint
+    }
+
+    /// Returns the chunk's byte data.
     pub fn payload(&self) -> &[u8] {
         &self.payload
     }
@@ -141,6 +148,29 @@ impl Chunk {
         Ok(Chunk { fprint, payload })
     }
 
+    /// Returns a chunk (potentially cached) from `payload`.
+    pub(crate) fn arc_from_bytes(payload: &[u8]) -> Arc<Chunk> {
+        let fprint = fingerprint_file_chunk(payload);
+
+        if let Some(ret) = fetch_from_cache(fprint) {
+            return ret;
+        }
+
+        let mut map = memmap2::MmapMut::map_anon(payload.len()).unwrap();
+
+        map.as_mut().copy_from_slice(payload);
+        let ret = Arc::new(
+            Chunk::new(
+                fingerprint_file_chunk(payload),
+                map.make_read_only().unwrap(),
+            )
+            .expect("must build: fingerprint matches"),
+        );
+
+        update_cache(ret.clone());
+        ret
+    }
+
     /// Returns a fresh chunk for `fprint`.
     #[cfg(test)]
     pub(crate) fn new_from_bytes(payload: &[u8]) -> Chunk {
@@ -404,31 +434,13 @@ fn maybe_decompress(payload: Vec<u8>, fprint: Fingerprint) -> Result<Vec<u8>> {
         }
     };
 
-    let decompress = |payload: &[u8]| -> std::io::Result<Vec<u8>> {
-        use std::io::Read;
-
-        let mut decoder = zstd::Decoder::new(&*payload)?;
-        let mut decompressed = vec![0; DECODED_CHUNK_SIZE_LIMIT];
-
-        match decoder.read(&mut decompressed) {
-            Ok(n) if n < DECODED_CHUNK_SIZE_LIMIT => {
-                decompressed.resize(n, 0u8);
-                decompressed.shrink_to_fit();
-                Ok(decompressed)
-            }
-            Ok(_) => Err(std::io::Error::new(
-                ErrorKind::Other,
-                "decoded zstd data >= DECODED_CHUNK_SIZE_LIMIT",
-            )),
-            Err(e) => Err(e),
-        }
+    let decompression_result = if let Some(r) = try_to_unzstd(&payload, DECODED_CHUNK_SIZE_LIMIT) {
+        r
+    } else {
+        return Ok(payload);
     };
 
-    if !payload.starts_with(&ZSTD_MAGIC) {
-        return Ok(payload);
-    }
-
-    match decompress(&payload) {
+    match decompression_result {
         Ok(decompressed) => {
             match compare_hash(&decompressed) {
                 Ok(()) => Ok(decompressed),
