@@ -1,6 +1,7 @@
 //! Replicated sqlite DBs are represented as protobuf "manifest"
 //! metadata that refer to content-addressed chunks by fingerprint.
 use std::fs::File;
+use std::sync::Arc;
 
 use tracing::instrument;
 use umash::Fingerprint;
@@ -10,9 +11,19 @@ use crate::chain_error;
 use crate::drop_result;
 use crate::fresh_error;
 use crate::fresh_warn;
+use crate::loader::Chunk;
+use crate::loader::Loader;
+use crate::replication_target::ReplicationTarget;
 use crate::result::Result;
+use crate::unzstd::try_to_unzstd;
 use crate::warn_from_os;
 
+/// Size limit for manifests after decompression.
+///
+/// Even a 1TB database fits its manifest in slightly more than 256 MB
+/// (2**28 bytes), so this ought to be large enough.
+const MANIFEST_SIZE_LIMIT: usize = 3usize << 27;
+
 /// A umash fingerprint.
 #[derive(Clone, PartialEq, Eq, prost::Message)]
 pub struct Fprint {
@@ -89,9 +100,34 @@ pub struct ManifestV1 {
     #[prost(sfixed32, tag = "7")]
     pub ctime_ns: i32,
 
+    // The fingerprint for the base array of bytes in `chunks`.
+    // Chunks has the correct length, but must be xor-ed with the
+    // little-endian contents of the data in `base_chunks_fprint` in
+    // order to find the real data.
+    //
+    // The base chunk, if provided, must consist of an even number of
+    // u64 words, in little endian.  Its length is otherwise
+    // arbitrary.  When shorter than the `chunks` array in the
+    // manifest, the base chunk is implicitly zero-padded (i.e, extra
+    // values in `chunks` are left untouched).  When the base chunk
+    // instead has more values than the `chunks` array, these extra
+    // u64 words are instead ignored.
+    //
+    // When generated by `Manifest::decode_and_validate`, any base
+    // chunk is always applied and this field is `None`.
+    #[prost(message, tag = "8")]
+    pub base_chunks_fprint: Option<Fprint>,
+
     // The fingerprints for each chunk as pairs of u64.  The first
     // chunk has fingerprint `chunks[0], chunks[1]`, the second
     // `chunks[2], chunks[3]`, etc.
+    //
+    // When generated by `Manifest::decode_and_validate`, this
+    // list of chunks always has `base_chunks_fprint` applied.
+    //
+    // If a base chunk has been applied to the `chunks` vector,
+    // `base_chunks_fprint` should always be populated; it's best to
+    // do so just before serialising the manifest.
     #[prost(fixed64, repeated, tag = "15")]
     pub chunks: Vec<u64>,
 }
@@ -103,18 +139,58 @@ pub struct ManifestV1 {
 pub struct Manifest {
     #[prost(message, tag = "1")]
     pub v1: Option<ManifestV1>,
+    // If we ever have a field #5 in this toplevel `Manifest` message,
+    // it must not be a varint: that could result in protobuf bytes
+    // that match the zstd magic header (whose first byte is 0x28).
+    //
+    // In fact, we might want to consider marking 5 as reserved, or
+    // guarantee we'll always zstd-compress manifest blob contents.
 }
 
 impl Manifest {
-    /// Attempts to decode the protobuf bytes in `buf` and
-    /// rejects clearly invalid data.
-    pub fn decode_and_validate(buf: &[u8], description: impl std::fmt::Debug) -> Result<Manifest> {
+    /// Attempts to decode the protobuf bytes in `buf`, potentially
+    /// after zstd-decompression, and rejects clearly invalid data.
+    ///
+    /// Uses the `targets` to fetch the base chunk if any is defined
+    /// in the return `Manifest`.  If there is such a chunk, it will
+    /// be fetched and applied to the return value.  The returned
+    /// manifest never has a `base_chunks_fprint`.
+    ///
+    /// When `targets` is `None`, it defaults to the globally configured
+    /// target list.
+    ///
+    /// If the original `Manifest` had defined a `base_chunks_fprint`,
+    /// the corresponding fingerprint and chunk bytes will be returned
+    /// in the second return value.
+    pub fn decode_and_validate(
+        buf: &[u8],
+        cache_builder: kismet_cache::CacheBuilder,
+        targets: Option<&[ReplicationTarget]>,
+        description: impl std::fmt::Debug,
+    ) -> Result<(Manifest, Option<Arc<Chunk>>)> {
         use prost::Message;
+        use std::convert::TryInto;
 
-        let manifest = Manifest::decode(buf)
-            .map_err(|e| chain_error!(e, "failed to parse proto manifest", ?description))?;
+        let try_parse = |buf: &[u8]| -> Result<Manifest> {
+            Manifest::decode(buf)
+                .map_err(|e| chain_error!(e, "failed to parse proto manifest", ?description))
+        };
 
-        let v1 = match &manifest.v1 {
+        let mut manifest = match try_to_unzstd(buf, MANIFEST_SIZE_LIMIT) {
+            None => try_parse(buf)?,
+            Some(Err(e)) => {
+                if let Ok(manifest) = try_parse(buf) {
+                    manifest
+                } else {
+                    return Err(chain_error!(e, "failed to unzstd manifest", ?description));
+                }
+            }
+            Some(Ok(decoded)) => try_parse(&decoded)?,
+        };
+
+        let mut base_chunk = None;
+
+        let v1 = match &mut manifest.v1 {
             Some(v1) => v1,
             None => return Err(fresh_error!("v1 format not found", ?manifest)),
         };
@@ -127,6 +203,54 @@ impl Manifest {
             return Err(fresh_error!("missing header_fprint", ?manifest));
         }
 
+        if let Some(base_fprint) = v1
+            .base_chunks_fprint
+            .as_ref()
+            .map(Into::<Fingerprint>::into)
+        {
+            let default_targets;
+
+            let targets = if let Some(targets) = targets {
+                targets
+            } else {
+                default_targets = crate::replication_target::get_default_replication_targets();
+                &default_targets.replication_targets
+            };
+            let loader = Loader::new(cache_builder, targets)?;
+
+            let chunk = if let Some(chunk) = loader.fetch_chunk(base_fprint)? {
+                chunk
+            } else {
+                return Err(fresh_error!("base fprint chunk not found", ?manifest));
+            };
+
+            // The base chunk should represent an array of
+            // fingerprints, and fingerprints span 16 bytes each.
+            if chunk.payload().len() % 16 != 0 {
+                return Err(fresh_error!(
+                    "base fprint chunk has invalid size",
+                    size = chunk.payload().len(),
+                    ?manifest
+                ));
+            }
+
+            let base_words = chunk
+                .payload()
+                .chunks_exact(std::mem::size_of::<u64>())
+                // Safe to unwrap: we get slices of exactly 8 bytes.
+                .map(|s| u64::from_le_bytes(s.try_into().unwrap()));
+
+            // `base_words` may contain an arbitrary number of
+            // little-endian u64 words: it's implicitly zero-padded if
+            // too short, and extra words are ignored.
+            for (word, base) in v1.chunks.iter_mut().zip(base_words) {
+                *word ^= base;
+            }
+
+            v1.base_chunks_fprint = None;
+            base_chunk = Some(chunk);
+        }
+
         // Check the chunk fingerprints: their fingerprint must match, and
         // there must be an even number of u64s (two per fingerprint).
         if Some(fingerprint_v1_chunk_list(&v1.chunks).into()) != v1.contents_fprint
@@ -135,7 +259,15 @@ impl Manifest {
             return Err(fresh_error!("invalid chunk list", ?manifest));
         }
 
-        Ok(manifest)
+        // The manifest's `base_chunks_fprint` should be empty (and
+        // applied) before we return the decoded manifest.
+        {
+            let base_chunks = || manifest.v1.as_ref()?.base_chunks_fprint.as_ref();
+
+            assert_eq!(base_chunks(), None);
+        }
+
+        Ok((manifest, base_chunk))
     }
 }
 
@@ -485,14 +617,25 @@ pub(crate) fn extract_manifest_len(manifest: &Manifest) -> Result<u64> {
     }
 }
 
-/// Returns the list of chunks in the manifest proto at `Path`, or
-/// an empty list if there is no such file.
+/// Returns the list of chunks in the manifest proto at `Path`, along
+/// with the `base_chunks_fprint` if any, or an empty list if there is
+/// no such file.
 #[instrument]
-pub(crate) fn parse_manifest_chunks(path: &std::path::Path) -> Result<Vec<Fingerprint>> {
+pub(crate) fn parse_manifest_chunks(
+    path: &std::path::Path,
+    replication_targets: &[ReplicationTarget],
+) -> Result<(Vec<Fingerprint>, Option<Arc<Chunk>>)> {
     match std::fs::read(path) {
-        Ok(contents) => extract_manifest_chunks(&Manifest::decode_and_validate(&*contents, path)?),
-
-        Err(e) if e.kind() == std::io::ErrorKind::NotFound => Ok(Vec::new()),
+        Ok(contents) => {
+            let (manifest, base_or) = Manifest::decode_and_validate(
+                &*contents,
+                Default::default(),
+                Some(replication_targets),
+                path,
+            )?;
+            Ok((extract_manifest_chunks(&manifest)?, base_or))
+        }
+        Err(e) if e.kind() == std::io::ErrorKind::NotFound => Ok((Vec::new(), None)),
         Err(e) => Err(chain_error!(e, "failed to open manifest proto file", ?path)),
     }
 }
