@@ -147,6 +147,7 @@ use std::io::ErrorKind;
 use std::os::unix::fs::PermissionsExt;
 use std::path::Path;
 use std::path::PathBuf;
+use std::sync::Arc;
 use std::sync::RwLock;
 use std::time::Duration;
 use tempfile::TempDir;
@@ -162,9 +163,11 @@ use crate::filtered_io_error;
 use crate::fresh_error;
 use crate::fresh_info;
 use crate::instance_id;
+use crate::loader::Chunk;
 use crate::manifest_schema::Manifest;
 use crate::ofd_lock::OfdLock;
 use crate::process_id::process_id;
+use crate::replication_target::ReplicationTarget;
 use crate::replication_target::ReplicationTargetList;
 use crate::result::Result;
 
@@ -863,15 +866,53 @@ pub(crate) fn chunk_name_fingerprint(name: &str) -> Option<Fingerprint> {
 
 /// Attempts to read a valid Manifest message from `file_path`.
 /// Returns Ok(None) if the file does not exist.
+///
+/// Any chunk necessary to decode the Manifest will be fetched from
+/// a cache built with the `cache_builder` and from the `targets`.
 #[instrument(level = "trace", err)]
-fn read_manifest_at_path(file_path: &Path) -> Result<Option<Manifest>> {
-    let contents = match std::fs::read(file_path) {
-        Ok(contents) => contents,
+fn read_manifest_at_path(
+    file_path: &Path,
+    cache_builder: kismet_cache::CacheBuilder,
+    targets: &[ReplicationTarget],
+) -> Result<Option<(Manifest, Option<Arc<Chunk>>)>> {
+    use std::io::Read;
+
+    let mut file = match std::fs::File::open(file_path) {
+        Ok(file) => file,
         Err(e) if e.kind() == ErrorKind::NotFound => return Ok(None),
-        Err(e) => return Err(chain_error!(e, "failed to read manifest", ?file_path)),
+        Err(e) => return Err(chain_error!(e, "failed to open manifest", ?file_path)),
     };
 
-    Ok(Some(Manifest::decode_and_validate(&*contents, file_path)?))
+    let mut contents = Vec::new();
+    file.read_to_end(&mut contents)
+        .map_err(|e| chain_error!(e, "failed to read manifest", ?file_path))?;
+
+    match Manifest::decode_and_validate(&*contents, cache_builder, Some(targets), file_path) {
+        Ok(ret) => Ok(Some(ret)),
+        Err(e) => {
+            use std::os::unix::fs::MetadataExt;
+
+            // If we failed to parse the manifest, but the manifest
+            // file doesn't exist anymore or was overwritten, bubble
+            // that up as a missing manifest file.
+            let initial_meta = file
+                .metadata()
+                .map_err(|e| chain_error!(e, "failed to stat open manifest file", ?file_path))?;
+
+            let current_meta = match std::fs::metadata(file_path) {
+                Ok(meta) => meta,
+                Err(e) if e.kind() == ErrorKind::NotFound => return Ok(None),
+                Err(e) => return Err(chain_error!(e, "failed to stat manifest file", ?file_path)),
+            };
+
+            if (initial_meta.dev(), initial_meta.ino()) == (current_meta.dev(), current_meta.ino())
+            {
+                Err(e)
+            } else {
+                Ok(None)
+            }
+        }
+    }
 }
 
 /// Returns the path to the staging directory under `parent`; it may
@@ -1174,36 +1215,76 @@ impl ReplicationBuffer {
         Ok(())
     }
 
-    /// Attempts to parse the current "consuming" manifest file.
+    /// Returns true if there is definitely a ready manifest file for
+    /// `db_path`.
+    pub fn has_ready_manifest(&self, db_path: &Path) -> bool {
+        let src = || -> Result<PathBuf> {
+            let mut src = self.spooling_directory.clone();
+            src.push(READY);
+            src = append_subdirectory(src);
+            src.push(META);
+            src.push(&percent_encode_local_path_uri(db_path)?);
+
+            Ok(src)
+        };
+
+        match src() {
+            Ok(src) => std::fs::symlink_metadata(&src).is_ok(),
+            Err(_) => false,
+        }
+    }
+
+    /// Attempts to parse the current "consuming" manifest file, and
+    /// returns the current base chunk for the list of chunk
+    /// fingerprints, if any.
     #[instrument(level = "trace", err)]
-    pub fn read_consuming_manifest(&self, db_path: &Path) -> Result<Option<Manifest>> {
+    pub fn read_consuming_manifest(
+        &self,
+        db_path: &Path,
+        cache_builder: kismet_cache::CacheBuilder,
+        targets: &[ReplicationTarget],
+    ) -> Result<Option<(Manifest, Option<Arc<Chunk>>)>> {
         let mut src = self.spooling_directory.clone();
         src.push(CONSUMING);
         src = append_subdirectory(src);
         src.push(META);
         src.push(&percent_encode_local_path_uri(db_path)?);
-        read_manifest_at_path(&src)
+        read_manifest_at_path(&src, cache_builder, targets)
     }
 
-    /// Attempts to parse the current ready manifest file.
+    /// Attempts to parse the current ready manifest file, and
+    /// returns the current base chunk for the list of chunk
+    /// fingerprints, if any.
     #[instrument(level = "trace", err)]
-    pub fn read_ready_manifest(&self, db_path: &Path) -> Result<Option<Manifest>> {
+    pub fn read_ready_manifest(
+        &self,
+        db_path: &Path,
+        cache_builder: kismet_cache::CacheBuilder,
+        targets: &[ReplicationTarget],
+    ) -> Result<Option<(Manifest, Option<Arc<Chunk>>)>> {
         let mut src = self.spooling_directory.clone();
         src.push(READY);
         src = append_subdirectory(src);
         src.push(META);
         src.push(&percent_encode_local_path_uri(db_path)?);
-        read_manifest_at_path(&src)
+        read_manifest_at_path(&src, cache_builder, targets)
     }
 
-    /// Attempts to parse the current staged manifest file.
+    /// Attempts to parse the current staged manifest file, and
+    /// returns the current base chunk for the list of chunk
+    /// fingerprints, if any.
     #[instrument(level = "trace", err)]
-    pub fn read_staged_manifest(&self, db_path: &Path) -> Result<Option<Manifest>> {
+    pub fn read_staged_manifest(
+        &self,
+        db_path: &Path,
+        cache_builder: kismet_cache::CacheBuilder,
+        targets: &[ReplicationTarget],
+    ) -> Result<Option<(Manifest, Option<Arc<Chunk>>)>> {
         let mut src = self.spooling_directory.clone();
         src.push(STAGING);
         src.push(META);
         src.push(&percent_encode_local_path_uri(db_path)?);
-        read_manifest_at_path(&src)
+        read_manifest_at_path(&src, cache_builder, targets)
     }
 
     /// Returns the path prefix for staged chunks.
