@@ -22,6 +22,7 @@ use crate::copier::Copier;
 use crate::drop_result;
 use crate::fresh_error;
 use crate::fresh_warn;
+use crate::loader::Chunk;
 use crate::manifest_schema::clear_version_id;
 use crate::manifest_schema::extract_version_id;
 use crate::manifest_schema::fingerprint_file_chunk;
@@ -37,6 +38,10 @@ use crate::result::Result;
 /// We snapshot db files in 64KB content-addressed chunks.
 pub(crate) const SNAPSHOT_GRANULARITY: u64 = 1 << 16;
 
+/// Don't generate a base fingerprint chunk for a list of fingerprints
+/// shorter than `BASE_CHUNK_MIN_LENGTH`.
+const BASE_CHUNK_MIN_LENGTH: usize = 1024;
+
 #[derive(Clone, Copy, Debug, Eq, PartialEq)]
 enum MutationState {
     Clean,   // No mutation since the last snapshot
@@ -80,6 +85,11 @@ pub(crate) struct Tracker {
     // The version id we observed at the beginning of the write
     // transaction, if any.
     previous_version_id: Vec<u8>,
+
+    // A shared mutable cell to retain the most recent base chunk for
+    // this db's manifest.  Keeping this chunk alive guarantees we can
+    // find it in the global cache, and thus avoids useless GETs.
+    recent_base_chunk: Option<Arc<crate::loader::Chunk>>,
 }
 
 fn flatten_chunk_fprints(fprints: &[Fingerprint]) -> Vec<u64> {
@@ -92,6 +102,126 @@ fn flatten_chunk_fprints(fprints: &[Fingerprint]) -> Vec<u64> {
     ret
 }
 
+/// What should we do with the current base chunk fingerprint list?
+enum BaseChunkAction {
+    /// Don't use a base chunk at all.
+    None,
+    /// Reuse the current base chunk.
+    Reuse,
+    /// Generate a new base chunk.
+    Refresh,
+}
+
+impl BaseChunkAction {
+    /// Determines what to do with a previous inherited base chunk
+    /// list.  The `matches` parameter counts the number of
+    /// fingerprints that match between the current chunk fingerprint
+    /// list and the inherited base chunk (i.e., the number of 0
+    /// fingerprints once the two are xor-ed together), and
+    /// `total_size` is the size of the current chunk fingerprint
+    /// list.
+    fn decide(matches: usize, total_size: usize) -> Self {
+        use rand::Rng;
+        let mut rng = rand::thread_rng();
+
+        // With this feature, we always return a random action to
+        // exercise the rest of the snapshotting code.
+        #[cfg(feature = "test_random_chunk_action")]
+        return match rng.gen_range(0..3) {
+            0 => BaseChunkAction::None,
+            1 => BaseChunkAction::Reuse,
+            _ => BaseChunkAction::Refresh,
+        };
+
+        // If there are enough matches that the random uniform integer
+        // is less than the number of matches, keep the old base chunk.
+        //
+        // We sample from an inclusive range to guarantee there's a
+        // small chance of recreating the base chunk even when
+        // `matches == total_size`.
+        if rng.gen_range(0..=total_size) < matches {
+            BaseChunkAction::Reuse
+        } else if total_size < BASE_CHUNK_MIN_LENGTH {
+            // The list of chunk fingerprints is too short to warrant
+            // a base chunk.  Regenerate to nothing.
+            BaseChunkAction::None
+        } else {
+            BaseChunkAction::Refresh
+        }
+    }
+}
+
+/// Re-encodes `chunks` to expose a more compressible list of `u64`s
+/// in the manifest.
+///
+/// On entry, `chunks` contains the real chunk fingerprints (before
+/// any xor-ing).  The return value holds the chunk fingerprints,
+/// potentially xor-ed with a base chunk.  If a base chunk was xor-ed
+/// in, it's the second return value.
+fn reencode_flattened_chunks(
+    repl: &ReplicationBuffer,
+    prev_base: Option<Arc<Chunk>>,
+    mut chunks: Vec<u64>,
+) -> Result<(Vec<u64>, Option<Arc<Chunk>>)> {
+    use std::convert::TryInto;
+
+    let apply_prev_base = |chunks: &mut Vec<u64>| {
+        let mut zero = 0;
+        // Apply `prev_base`, the previous base chunk, to see what's left.
+        if let Some(prev) = prev_base.as_ref() {
+            let base_words = prev
+                .payload()
+                .chunks_exact(std::mem::size_of::<u64>())
+                // Safe to unwrap: we get slices of exactly 8 bytes.
+                .map(|s| u64::from_le_bytes(s.try_into().unwrap()));
+
+            for (word, base) in chunks.iter_mut().zip(base_words) {
+                let result = *word ^ base;
+                *word = result;
+
+                zero += (result == 0) as usize;
+            }
+        }
+
+        zero
+    };
+
+    let matches = apply_prev_base(&mut chunks);
+    match BaseChunkAction::decide(matches, chunks.len()) {
+        BaseChunkAction::None => {
+            // Apply `prev_base` again to undo the xors.
+            apply_prev_base(&mut chunks);
+            return Ok((chunks, None));
+        }
+        BaseChunkAction::Reuse => return Ok((chunks, prev_base)),
+        BaseChunkAction::Refresh => {}
+    }
+
+    // Copy the original value of `chunks` in `base`, as little-endian
+    // bytes.
+    let mut base: Vec<u8> = Vec::new();
+
+    base.reserve(chunks.len() * std::mem::size_of::<u64>());
+    for chunk in chunks.iter() {
+        base.extend(chunk.to_le_bytes());
+    }
+
+    // And now un-apply `prev_base`.
+    if let Some(prev) = prev_base {
+        for (base, prev) in base.iter_mut().zip(prev.payload()) {
+            *base ^= prev;
+        }
+    }
+
+    chunks.fill(0);
+
+    let base_chunk = Chunk::arc_from_bytes(&base);
+    std::mem::drop(base);
+
+    repl.stage_chunk(base_chunk.fprint(), base_chunk.payload())?;
+    Ok((chunks, Some(base_chunk)))
+}
+
 fn rebuild_chunk_fprints(flattened: &[u64]) -> Option<Vec<Fingerprint>> {
     if (flattened.len() % 2) != 0 {
         return None;
@@ -164,6 +294,7 @@ impl Tracker {
             dirty_chunks: BTreeMap::new(),
             backing_file_state: MutationState::Unknown,
             previous_version_id: Vec::new(),
+            recent_base_chunk: None,
         }))
     }
 
@@ -463,12 +594,35 @@ impl Tracker {
         Ok((len, chunk_fprints, num_snapshotted))
     }
 
+    /// Reads the current staged manifest, if any.
+    fn read_current_manifest(&self) -> Result<Option<(Manifest, Option<Arc<Chunk>>)>> {
+        let buf = &self.buffer;
+        let mut builder = kismet_cache::CacheBuilder::new();
+
+        // We assume any chunk needed to parse the manifest is always
+        // available in the staged chunk directory, and make sure to
+        // save such chunks during GC.
+        //
+        // We do that instead of using the test-only
+        // `cache_builder_for_source` because the latter does a lot
+        // more work in order to look into `ready` and `consuming`
+        // subdirectories for data chunk.  However, we (should) always
+        // keep the current base chunk in `staged`, so that extra work
+        // is useless and could hide logic bugs.
+        builder.plain_reader(buf.staged_chunk_directory());
+        buf.read_staged_manifest(
+            &self.path,
+            builder,
+            &self.replication_targets.replication_targets,
+        )
+    }
+
     /// Snapshots the contents of the tracked file to its replication
     /// buffer.  Concurrent threads or processes may be doing the same,
     /// but the contents of the file can't change, since we still hold
     /// a sqlite read lock on the db file.
     #[instrument(skip(self), err)]
-    fn snapshot_file_contents(&self) -> Result<()> {
+    fn snapshot_file_contents(&mut self) -> Result<()> {
         use std::os::unix::fs::MetadataExt;
 
         let buf = &self.buffer;
@@ -498,8 +652,8 @@ impl Tracker {
                          e => chain_warn!(e, "failed to force populate version xattr", path=?self.path));
         }
 
-        let mut current_manifest: Option<Manifest> = buf
-            .read_staged_manifest(&self.path)
+        let mut current_manifest: Option<(Manifest, Option<Arc<Chunk>>)> = self
+            .read_current_manifest()
             .map_err(|e| chain_info!(e, "failed to read staged manifest file"))
             .ok()
             .flatten();
@@ -516,7 +670,7 @@ impl Tracker {
         if self.backing_file_state == MutationState::Dirty {
             let mut up_to_date = false;
 
-            if let Some(manifest) = &current_manifest {
+            if let Some((manifest, _)) = &current_manifest {
                 if let Some(v1) = &manifest.v1 {
                     // The current snapshot seems to have everything
                     // up to our last write transaction.  We can use that!
@@ -572,7 +726,7 @@ impl Tracker {
             // to do.  We don't even have to update "ready": the `Copier`
             // will read from the staging directory when we don't change it.
             if !version_id.is_empty() {
-                if let Some(manifest) = &current_manifest {
+                if let Some((manifest, _)) = &current_manifest {
                     if matches!(&manifest.v1, Some(v1) if v1.version_id == version_id) {
                         return Ok(());
                     }
@@ -595,14 +749,21 @@ impl Tracker {
 
         // Try to get an initial list of chunks to work off.
         let mut base_fprints = None;
-        if let Some(manifest) = current_manifest {
-            if let Some(v1) = manifest.v1 {
+        if let Some((manifest, _)) = current_manifest.as_ref() {
+            if let Some(v1) = &manifest.v1 {
                 base_fprints = rebuild_chunk_fprints(&v1.chunks);
             }
         }
 
-        let (copied, chunks) = {
-            let (len, chunks, copied) = self.snapshot_chunks(base_fprints)?;
+        // `copied` and `chunks` always take the both
+        // newly-snapshotted chunks and the base chunk into account.
+        //
+        // The base chunk, if any, is also available on its own for
+        // full GCs after publishing: if we just published a new ready
+        // manifest, we can delete all chunks *except the current base
+        // chunk*.
+        let (copied, chunks, base_chunk) = {
+            let (len, mut chunks, mut copied) = self.snapshot_chunks(base_fprints)?;
 
             let (ctime, ctime_ns) = match self.file.metadata() {
                 Ok(meta) => (meta.ctime(), meta.ctime_nsec() as i32),
@@ -614,6 +775,14 @@ impl Tracker {
 
             let flattened = flatten_chunk_fprints(&chunks);
             let manifest_fprint = fingerprint_v1_chunk_list(&flattened);
+            let (compressible, base_chunk) = reencode_flattened_chunks(
+                &self.buffer,
+                current_manifest.map(|x| x.1).flatten(),
+                flattened,
+            )?;
+
+            let base_fprint = base_chunk.as_ref().map(|x| x.fprint());
+
             let manifest = Manifest {
                 v1: Some(ManifestV1 {
                     header_fprint: Some(header_fprint.into()),
@@ -622,22 +791,29 @@ impl Tracker {
                     len,
                     ctime,
                     ctime_ns,
-                    chunks: flattened,
+                    base_chunks_fprint: base_fprint.map(|fp| fp.into()),
+                    chunks: compressible,
                 }),
             };
 
             buf.publish_manifest(&self.path, &manifest)?;
 
-            (copied, chunks)
+            // We published our new manifest, so we'll probably find a
+            // reference to `base_chunk` the next time we load the
+            // staged manifest.  Let's make sure it remains cached.
+            self.recent_base_chunk = base_chunk;
+
+            if let Some(base_fprint) = base_fprint {
+                copied += 1;
+                chunks.push(base_fprint);
+            }
+
+            (copied, chunks, base_fprint)
         };
 
         let mut published = false;
-        // If we can publish a new ready manifest, try to do so.
-        if matches!(
-            buf.read_ready_manifest(&self.path)
-                .map_err(|e| chain_info!(e, "failed to read ready manifest", path=?self.path)),
-            Ok(None)
-        ) {
+        // Unless there obviously is a ready manifest, try to publish our own.
+        if !buf.has_ready_manifest(&self.path) {
             let ready = buf.prepare_ready_buffer(&chunks)?;
 
             published = buf
@@ -656,9 +832,11 @@ impl Tracker {
         // the copier that we only remove chunks after attempting
         // to publish the ready buffer.
         if published {
-            // If we just published our snapshot to the ready
-            // buffer, we can delete all chunks.
-            drop_result!(buf.gc_chunks(&[]),
+            // If we just published our snapshot to the ready buffer,
+            // we can delete all chunks... except the base chunk for
+            // the current manifest, which must remain readable.
+            let chunks = base_chunk.iter().copied().collect::<Vec<_>>();
+            drop_result!(buf.gc_chunks(&chunks),
                          e => chain_info!(e, "failed to clear all staged chunks", path=?self.path));
             self.chunk_counter.store(0, Ordering::Relaxed);
         } else {
@@ -785,11 +963,7 @@ enum ChunkSource {
 
 #[cfg(feature = "test_vfs")]
 impl Tracker {
-    fn fetch_snapshot_or_die(
-        &self,
-        manifest: &Manifest,
-        source: ChunkSource,
-    ) -> crate::snapshot::Snapshot {
+    fn cache_builder_for_source(&self, source: ChunkSource) -> kismet_cache::CacheBuilder {
         let buf = &self.buffer;
         let mut cache_builder = kismet_cache::CacheBuilder::new();
 
@@ -827,10 +1001,19 @@ impl Tracker {
             cache_builder.plain_reader(buf.consuming_chunk_directory());
         }
 
+        cache_builder
+    }
+
+    fn fetch_snapshot_or_die(
+        &self,
+        manifest: &Manifest,
+        source: ChunkSource,
+    ) -> crate::snapshot::Snapshot {
         crate::snapshot::Snapshot::new(
-            cache_builder,
+            self.cache_builder_for_source(source),
             &self.replication_targets.replication_targets,
             manifest,
+            None,
         )
         .expect("failed to instantiate snapshot")
     }
@@ -839,14 +1022,14 @@ impl Tracker {
     /// every chunk in that snapshot.
     fn validate_snapshot(
         &self,
-        manifest_or: crate::result::Result<Option<Manifest>>,
+        manifest_or: crate::result::Result<Option<(Manifest, Option<Arc<Chunk>>)>>,
         source: ChunkSource,
     ) -> Result<()> {
         let manifest = match manifest_or {
             // If the manifest file can't be found, assume it was
             // replicated correctly, and checked earlier.
             Ok(None) => return Ok(()),
-            Ok(Some(manifest)) => manifest,
+            Ok(Some((manifest, _))) => manifest,
             Err(err) => return Err(err),
         };
 
@@ -859,14 +1042,28 @@ impl Tracker {
     #[cfg(feature = "test_validate_reads")]
     fn validate_all_snapshots(&self) {
         let buf = &self.buffer;
+        let targets = &self.replication_targets.replication_targets;
+
         self.validate_snapshot(
-            buf.read_consuming_manifest(&self.path),
+            buf.read_consuming_manifest(
+                &self.path,
+                self.cache_builder_for_source(ChunkSource::Consuming),
+                targets,
+            ),
             ChunkSource::Consuming,
         )
         .expect("consuming snapshot must be valid");
-        self.validate_snapshot(buf.read_ready_manifest(&self.path), ChunkSource::Ready)
-            .expect("ready snapshot must be valid");
-        self.validate_snapshot(buf.read_staged_manifest(&self.path), ChunkSource::Staged)
+        self.validate_snapshot(
+            buf.read_ready_manifest(
+                &self.path,
+                self.cache_builder_for_source(ChunkSource::Ready),
+                targets,
+            ),
+            ChunkSource::Ready,
+        )
+        .expect("ready snapshot must be valid");
+
+        self.validate_snapshot(self.read_current_manifest(), ChunkSource::Staged)
             .expect("staged snapshot must be valid");
     }
 
@@ -876,14 +1073,26 @@ impl Tracker {
         use blake2b_simd::Params;
 
         let buf = &self.buffer;
+        let targets = &self.replication_targets.replication_targets;
 
         self.validate_snapshot(
-            buf.read_consuming_manifest(&self.path),
+            buf.read_consuming_manifest(
+                &self.path,
+                self.cache_builder_for_source(ChunkSource::Consuming),
+                targets,
+            ),
             ChunkSource::Consuming,
         )
         .expect("consuming snapshot must be valid");
-        self.validate_snapshot(buf.read_ready_manifest(&self.path), ChunkSource::Ready)
-            .expect("ready snapshot must be valid");
+        self.validate_snapshot(
+            buf.read_ready_manifest(
+                &self.path,
+                self.cache_builder_for_source(ChunkSource::Ready),
+                targets,
+            ),
+            ChunkSource::Ready,
+        )
+        .expect("ready snapshot must be valid");
 
         // The VFS layer doesn't do anything with the file's offset,
         // and all locking uses OFD locks, so this `dup(2)` is fine.
@@ -904,8 +1113,8 @@ impl Tracker {
         };
 
         let mut hasher = Params::new().hash_length(32).to_state();
-        let manifest = buf
-            .read_staged_manifest(&self.path)
+        let (manifest, _) = self
+            .read_current_manifest()
             .expect("manifest must parse")
             .expect("manifest must exist");
 
